@@ -1,4 +1,6 @@
+import sys
+
 from find_libpython import main
 
 if __name__ == "__main__":
-    main()
+    sys.exit(main())
